package massim.agent.framework;

import java.util.concurrent.LinkedBlockingQueue;

public abstract class AbstractSyncAgent extends AbstractAsyncAgent {

    private LinkedBlockingQueue<MessageContent> inbox;
    private Perception perception;
    private String action;
    private Boolean bActed = false;

    public AbstractSyncAgent(String username, String password) {
        super(username, password);
        inbox = new LinkedBlockingQueue<MessageContent>();
        perception = null;
    }

    protected final String processRequestAction(Perception perception) {
        synchronized (this) {
            this.perception = perception;
            action = "skip";
            bActed = false;
        }

        // wait some time for the agent to act
        // this has to remain unlocked in order to allow the agent to reason and act
        // we wait until the agent calles act(), but at most 4*450ms=1800ms
        String returnAction = "skip";
        try {
            Boolean acted = false;
            int wakeUps = 0;
            while ((!acted) && (wakeUps <= 4)) {
                wakeUps++;
                Thread.sleep(450);
                // read the acted flag with caution in a synchronized mode!
                synchronized (this) {
                    acted = bActed;
                    returnAction = action;
                }
            }
        } catch (InterruptedException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        synchronized (this) {
            action = "skip";
        }

        // return the action which was set by the agent
        return returnAction;
    }

    /**
     * Retrieves a message from the communication inbox.
     *
     * Watch out! The method is blocking, i.e., in the case there is no message in the inbox
     * it waits for a message to arrive.
     */
    public synchronized final Perception sense() {
        return perception;
    }

    /**
     * Returns true if the communication inbox is empty.
     *
     * The method is non-blocking. I.e., it returns immediately regardless of whether
     * there are messages in the message queue, or not.
     */
    public synchronized final Boolean isEmptySensor() {
        return (perception == null);
    }

    /**
     *
     */
    public synchronized final void resetSensor() {
        perception = null;
    }

    /**
     *  Message arrival handler for the AsynchronousAgent class
     */
    @Override
    protected synchronized final void onMessageReceivedPrivate(MessageContent content) {
        onMessageReceivedToInbox(content);
    }

    /**
     *  The handler of message arrival to the inbox
     * @param message
     */
    private synchronized final void onMessageReceivedToInbox(MessageContent message) {
        try {
            inbox.put(message);
        } catch (InterruptedException e) {
            System.err.println("ERROR: Inbox handling corrupted!");
            e.printStackTrace();
        }
    }

    /**
     * Retrieves a message from the communication inbox.
     *
     * Watch out! The method is blocking, i.e., in the case there is no message in the inbox
     * it waits for a message to arrive.
     */
    public synchronized final MessageContent popInboxMessage() {
        try {
            return inbox.take();
        } catch (InterruptedException e) {
            System.err.println("ERROR: message queue popInboxMessage() was interrupted!");
            e.printStackTrace();
        }

        // something very bad happened!
        return null;
    }

    /**
     * Returns true if the communication inbox is empty.
     *
     * The method is non-blocking. I.e., it returns immediately regardless of whether
     * there are messages in the message queue, or not.
     */
    public synchronized final Boolean isEmptyInbox() {
        return inbox.isEmpty();
    }

    /**
     * Handler of inbox message receipt.
     */
    @Override
    protected synchronized final void onMessageReceived(MessageContent content) {
        System.err.println("Wrong use of AbstractSyncAgent class! Synchronous agent must actively poll its message queue in order to retrieve messages");
    }

    /**
     * DOES NOT CHECK VALIDTY OF THE ACTION!!!
     * @param action
     */
    public synchronized final void act(String action) {
        this.action = action;
        this.bActed = true;
    }

    /**
     * The launch method for the synchronous agent implementation.
     *
     *  Here the user can implement whatever continuous reasoning algorithms
     *  which then interacts with the simulation only when it wants and is
     *  never explicitely bothered by anybody by callbacks.
     */
    protected abstract void runSyncAgent();
}
